from ._VersionInformation import VersionInfomation
from ._PersonInfomation import PersonInfomation
from ._AuthorInfomation import AuthorInfomation
from ._Rect import IRect, GenericRect, Rect, DiscreteRect
